import { Component } from '@angular/core';
import {
  ColDef,
  CsvExportParams,
  GridApi,
  GridReadyEvent,
  ProcessCellForExportParams,
  ValueGetterParams,
} from 'ag-grid-community';
import { Employer } from '../../../models/employer.model';
import { Job } from '../../../models/job.model';
import { Structure } from '../../../models/structure.model';
import { TempUser } from '../../../models/temp-user.model';
import { User } from '../../../models/user.model';
import { NotificationService } from '../../../services/notification.service';
import { TempUserService } from '../../../services/temp-user.service';
import { AdminService } from '../../services/admin.service';
import { BooleanCellRendererComponent } from '../boolean-cell-renderer/boolean-cell-renderer';
import { AdministredStructuresComponent } from './administred-structures/administred-structures.component';
import { DeleteUserComponent } from './delete-user/delete-user.component';

type UsersStatus = 'attached' | 'unattached';

@Component({
  selector: 'app-admin-manage-users',
  templateUrl: './manage-users.component.html',
  styleUrls: ['../../admin.scss'],
})
export class ManageUsersComponent {
  public attachedUsers: User[] = [];
  public gridApiAttached: GridApi;
  public unAttachedUsers: User[] = [];
  public gridApiUnattached: GridApi;
  public unVerifiedUsers: User[] = [];
  public validatedJobsNames: string[] = [];
  public validatedJobs: Job[] = [];
  public validatedEmployersNames: string[] = [];
  public validatedEmployers: Employer[] = [];
  public deleteModalOpened = false;
  public editJobModalOpened = false;
  public editEmployerModalOpened = false;
  public userToDelete: User = null;

  public tempUsers: TempUser[];

  // Date handling methods
  public dateValueGetter = (params: any): Date | null => {
    const value: unknown = params.data[params.colDef.field];
    if (typeof value === 'string') {
      const [day, month, year] = value.split('/').map(Number);
      if (day && month && year) {
        return new Date(year, month - 1, day);
      }
    }
    return null;
  };
  public dateValueFormatter = (params: any): string => {
    return params.value instanceof Date ? params.value.toLocaleDateString('fr-FR') : '';
  };
  public dateComparator = (dateA: Date, dateB: Date): number => {
    return (dateA?.getTime() ?? 0) - (dateB?.getTime() ?? 0);
  };

  public changingJobs: Job[] = [
    { name: null, hasPersonalOffer: true, validated: false },
    { name: null, hasPersonalOffer: true, validated: false },
  ];
  public changingEmployers: Employer[] = [
    {
      name: null,
      validated: false,
      users: [],
    },
    {
      name: null,
      validated: false,
      users: [],
    },
  ];
  public contextRow: any;
  public components;
  public defaultColDef: ColDef = {
    editable: true,
    sortable: true,
    flex: 1,
    minWidth: 100,
    filter: true,
    resizable: true,
  };
  public unAttachedColDef: ColDef = {
    ...this.defaultColDef,
    cellStyle: {
      color: '#da6c2e',
    },
  };

  public columnDefsUnattached: ColDef<User>[];
  public columnDefs: ColDef<User>[] = [
    {
      headerName: 'Nom',
      field: 'surname',
    },
    {
      headerName: 'Prénom',
      field: 'name',
    },
    {
      minWidth: 150,
      headerName: 'Mail',
      field: 'email',
    },
    {
      minWidth: 100,
      headerName: 'Téléphone',
      field: 'phone',
    },
    {
      headerName: 'Structures administrées',
      cellRenderer: 'administredStructuresComponent',
      // valueFormatter: (params) => this.formatStructureCell(params.value),
      cellRendererParams: {
        structures: 'structures',
      },
      minWidth: 350,
      editable: false,
    },
    {
      headerName: 'Fonction',
      field: 'job',
      cellEditor: 'agSelectCellEditor',
      singleClickEdit: true,
      cellEditorParams: (): { values: string[] } => {
        return {
          values: ['', ...this.validatedJobsNames],
        };
      },
      valueGetter: (params) => params.data.job?.name ?? '',
      valueSetter: (params): boolean => {
        params.data.job = this.validatedJobs.find((job) => job.name === params.newValue);
        return true;
      },
      onCellValueChanged: this.onJobChange.bind(this),
    },
    {
      headerName: 'Employeur',
      field: 'employer',
      cellEditor: 'agSelectCellEditor',
      singleClickEdit: true,
      cellEditorParams: (): { values: string[] } => {
        return {
          values: ['', ...this.validatedEmployersNames],
        };
      },
      valueGetter: (params) => params.data.employer?.name ?? '',
      valueSetter: (params): boolean => {
        params.data.employer = this.validatedEmployers.find((employer) => employer.name === params.newValue);
        return true;
      },
      onCellValueChanged: this.onEmployerChange.bind(this),
    },
    {
      headerName: 'Référent numérique',
      field: 'isReferent',
      cellRenderer: 'isReferent',
      cellEditor: 'agSelectCellEditor',
      valueGetter: this.translateBoolean,
      singleClickEdit: true,
      cellEditorParams: {
        values: [true, false],
        cellEditorPopup: true,
      },
      onCellValueChanged: this.toggleIsReferent.bind(this),
    },
    {
      headerName: 'Date de création',
      field: 'createdAt',
      minWidth: 150,
      sortable: true,
      valueGetter: this.dateValueGetter,
      valueFormatter: this.dateValueFormatter,
      comparator: this.dateComparator,
    },
    {
      headerName: 'Dernière connexion',
      field: 'lastLoginDate',
      minWidth: 150,
      sortable: true,
      valueGetter: this.dateValueGetter,
      valueFormatter: this.dateValueFormatter,
      comparator: this.dateComparator,
    },
    {
      headerName: 'Actions',
      editable: false,
      minWidth: 150,
      cellRenderer: 'deleteUserComponent',
      cellRendererParams: {
        onClick: this.onDeleteButtonClick.bind(this),
        label: 'Supprimer',
      },
      cellStyle: { 'text-align': 'center' },
    },
  ];

  public colDefsTempUsers: ColDef<TempUser>[] = [
    {
      minWidth: 100,
      headerName: 'E-mail',
      field: 'email',
    },
    {
      headerName: 'Structure(s)',
      cellRenderer: 'administredStructuresComponent',
      cellRendererParams: {
        structures: 'structures',
      },
      minWidth: 350,
      editable: false,
    },
    {
      minWidth: 50,
      headerName: 'Demande de rattachement envoyée le',
      field: 'updatedAt',
      valueGetter: this.dateValueGetter,
      valueFormatter: this.dateValueFormatter,
      comparator: this.dateComparator,
    },
  ];

  constructor(
    private adminService: AdminService,
    private tempUserService: TempUserService,
    private notificationService: NotificationService,
  ) {
    this.findValidatedJobs();
    this.findValidatedEmployers();
    this.columnDefsUnattached = [...this.columnDefs];
    this.columnDefsUnattached = this.columnDefsUnattached.filter((item) => {
      const headerName = String(item.headerName).toLowerCase();
      return !headerName.includes('structure');
    });

    this.columnDefsUnattached.splice(this.columnDefsUnattached.length - 1, 0, {
      headerName: 'Non rattaché depuis',
      field: 'unattachedSince' as any,
      minWidth: 180,
      valueGetter: this.dateValueGetter,
      valueFormatter: this.dateValueFormatter,
      comparator: this.dateComparator,
      editable: false,
    });
    this.components = {
      deleteUserComponent: DeleteUserComponent,
      isReferent: BooleanCellRendererComponent,
      administredStructuresComponent: AdministredStructuresComponent,
    };
    this.findAttachedUsers();
    this.findUnAttachedUsers();
    this.findUnVerifiedUsers();
    this.findTempUsers();
  }

  public onDeleteButtonClick(arg): void {
    this.deleteUser(arg.data, false);
  }

  public onJobChange(arg): void {
    const selectedJob = arg.data.job;
    const oldJob = new Job({ name: arg.oldValue });
    if (selectedJob === undefined && oldJob !== undefined) {
      this.toggleEditJobModal(null, oldJob, arg);
      return;
    }
    if (selectedJob && oldJob.name === '') {
      // user had no job assigned, and admin selected a job -> call setUserJob without modal
      this.adminService.setUserJob(selectedJob._id, arg.data.id).subscribe(() => {
        arg.node.data[arg.colDef.field] = selectedJob;
        arg.api.refreshCells({ rowNodes: [arg.node], columns: [arg.column.colId] });
        this.notificationService.showSuccess('La fonction a bien été attribuée.');
      });
      return;
    }
    if (selectedJob === oldJob || selectedJob.name === oldJob.name) {
      arg.node.data[arg.colDef.field] = selectedJob;
      arg.api.refreshCells({ rowNodes: [arg.node], columns: [arg.column.colId] });
      return;
    }
    this.toggleEditJobModal(selectedJob, oldJob, arg);
  }

  public onEmployerChange(arg): void {
    const selectedEmployer = arg.data.employer;
    const oldEmployer = new Employer({ name: arg.oldValue });
    if (selectedEmployer === undefined && oldEmployer !== undefined) {
      this.toggleEditEmployerModal(null, oldEmployer, arg);
      return;
    }
    if (selectedEmployer && oldEmployer.name === '') {
      // user had no employer assigned, and admin selected a employer -> call setUserEmployer
      this.adminService.setUserEmployer(selectedEmployer._id, arg.data.id).subscribe(() => {
        arg.node.data[arg.colDef.field] = selectedEmployer;
        arg.api.refreshCells({ rowNodes: [arg.node], columns: [arg.column.colId] });
        this.notificationService.showSuccess("L'employeur a bien été attribué.");
      });
      return;
    }
    if (selectedEmployer === oldEmployer || selectedEmployer.name === oldEmployer.name) {
      arg.node.data[arg.colDef.field] = selectedEmployer;
      arg.api.refreshCells({ rowNodes: [arg.node], columns: [arg.column.colId] });
      return;
    }
    this.toggleEditEmployerModal(selectedEmployer, oldEmployer, arg);
  }

  public deleteUser(user: User, shouldDelete: boolean): void {
    this.toggleDeleteModal(user);
    if (shouldDelete) {
      this.adminService.deleteUser(user._id).subscribe((data) => {
        this.unAttachedUsers = this.unAttachedUsers.filter((obj) => obj._id !== data._id);
        this.unVerifiedUsers = this.unVerifiedUsers.filter((obj) => obj._id !== data._id);
        this.attachedUsers = this.attachedUsers.filter((obj) => obj._id !== data._id);
      });
    }
  }

  public translateBoolean(params: ValueGetterParams): '✔️' | '❌' {
    return params.data.isReferent ? '✔️' : '❌';
  }

  public toggleIsReferent(arg): void {
    if (arg.oldValue === arg.newValue) {
      return;
    } else {
      this.adminService.toggleUserIsReferent(arg.data.id).subscribe(() => {
        this.notificationService.showSuccess("L'opération a réussie.");
      });
    }
  }

  public editJob(newJob: Job, oldJob: Job, shouldEdit: boolean, context?): void {
    this.toggleEditJobModal(newJob, oldJob, context);
    if (shouldEdit) {
      if (newJob) {
        // store newJob otherwise cell will only keep name value
        this.adminService.setUserJob(newJob._id, context.node.data._id).subscribe((data) => {
          context.node.data[context.colDef.field] = data.job;
          context.api.refreshCells({ rowNodes: [context.node], columns: [context.column.colId] });
          this.notificationService.showSuccess("L'opération a réussie.");
        });
      } else {
        this.adminService.removeUserJob(context.node.data._id).subscribe((data) => {
          context.node.data[context.colDef.field] = data.job;
          context.api.refreshCells({ rowNodes: [context.node], columns: [context.column.colId] });
          this.notificationService.showSuccess("L'opération a réussie.");
        });
      }
    } else {
      context.node.data[context.colDef.field] = new Job({ name: context.oldValue });
      context.api.refreshCells({ rowNodes: [context.node], columns: [context.column.colId], force: true });
    }
  }

  public editEmployer(newEmployer: Employer, oldEmployer: Employer, shouldEdit: boolean, context?): void {
    this.toggleEditEmployerModal(newEmployer, oldEmployer, context);
    if (shouldEdit) {
      if (newEmployer) {
        this.adminService.setUserEmployer(newEmployer._id, context.node.data._id).subscribe((data) => {
          context.node.data[context.colDef.field] = data.employer;
          context.api.refreshCells({ rowNodes: [context.node], columns: [context.column.colId] });
          this.notificationService.showSuccess("L'opération a réussie.");
        });
      } else {
        this.adminService.removeUserEmployer(context.node.data._id).subscribe((data) => {
          context.node.data[context.colDef.field] = data.employer;
          context.api.refreshCells({ rowNodes: [context.node], columns: [context.column.colId] });
          this.notificationService.showSuccess("L'opération a réussie.");
        });
      }
    } else {
      context.node.data[context.colDef.field] = new Employer({ name: context.oldValue });
      context.api.refreshCells({ rowNodes: [context.node], columns: [context.column.colId] });
    }
  }

  public toggleDeleteModal(userToDelete: User): void {
    this.userToDelete = userToDelete;
    this.deleteModalOpened = !this.deleteModalOpened;
  }

  public toggleEditJobModal(newValue: Job, oldValue: Job, dataGrid): void {
    this.changingJobs = [newValue, oldValue];
    this.contextRow = dataGrid;
    this.editJobModalOpened = !this.editJobModalOpened;
  }

  public toggleEditEmployerModal(newValue: Employer, oldValue: Employer, dataGrid): void {
    this.changingEmployers = [newValue, oldValue];
    this.contextRow = dataGrid;
    this.editEmployerModalOpened = !this.editEmployerModalOpened;
  }

  public findValidatedJobs(): void {
    this.adminService.getJobs().subscribe((jobs) => {
      this.validatedJobs = jobs;
      this.validatedJobsNames = jobs.map((job) => job.name);
    });
  }

  public findValidatedEmployers(): void {
    this.adminService.getEmployers().subscribe((employers) => {
      this.validatedEmployers = employers;
      this.validatedEmployersNames = employers.map((employer) => employer.name);
    });
  }

  public findAttachedUsers(): void {
    this.adminService.getAttachedUsers().subscribe((users) => {
      this.attachedUsers = users;
      this.attachedUsers.forEach((user) => {
        user._id = user.id;
      });
    });
  }

  public findUnAttachedUsers(): void {
    this.adminService.getUnAttachedUsers().subscribe((users) => {
      this.unAttachedUsers = users;
      this.unAttachedUsers.forEach((user) => {
        user._id = user.id;
      });
    });
  }

  public findUnVerifiedUsers(): void {
    this.adminService.getUnVerifiedUsers().subscribe((users) => {
      this.unVerifiedUsers = users;
      this.unVerifiedUsers.forEach((user) => {
        user._id = user.id;
      });
    });
  }

  public findTempUsers(): void {
    // TODO: Change server side so it returns structureNames instead of ids
    this.tempUserService.getAllTempUsers().subscribe((data) => {
      this.tempUsers = data;
    });
  }

  public getRowHeight(params): number {
    const structures = params.data.structures || [];
    return structures.length !== 0 ? structures.length * 40 : 40;
  }

  public onGridReady(params: GridReadyEvent, status: UsersStatus): void {
    if (status === 'attached') {
      this.gridApiAttached = params.api;
    } else if (status === 'unattached') {
      this.gridApiUnattached = params.api;
    }
  }

  public exportUsers(status: UsersStatus): void {
    const gridApi = status === 'attached' ? this.gridApiAttached : this.gridApiUnattached;

    if (!gridApi) {
      throw new Error('Grid API not found for the selected status.');
    }

    const exportParams: CsvExportParams = {
      columnKeys: ['name', 'surname', 'email', 'job', 'structures'],
      processCellCallback: (params) => this.formatCells(params),
      fileName: `RESIN - export utilisateurs ${status === 'attached' ? 'rattachés' : 'non rattachés'}`,
    };

    gridApi.exportDataAsCsv(exportParams);
  }

  private formatCells(params: ProcessCellForExportParams): string {
    const cellRenderer = params.column.getColDef().cellRenderer;

    if (!cellRenderer) {
      return this.formatRegularCell(params);
    }

    switch (cellRenderer) {
      case 'administredStructuresComponent':
        return this.formatStructureCell(params.value);
      default:
        return params.value;
    }
  }

  private formatRegularCell(params: any): string {
    return params?.value;
  }

  private formatStructureCell(params: any[]): string {
    if (params?.length > 0) {
      const structureNames = params.map((structure: Structure) => structure.structureName);
      return structureNames.join(', ');
    }
    return '';
  }
}
